import random
class A(object):
    def __init__(self, id=None) -> None:
        super().__init__()
        if id is None:
            id = random.randint(0, 1000000)
        self.id = id
    
    def quit(self):
        print(f"A [{self.id} is quiting]")

    def __repr__(self) -> str:
        return f"A [{self.id}]"

from threading import Thread
import threading
from queue import Queue

class Mgr(object):
    def __init__(self) -> None:
        super().__init__()
        self._items = {}
        self.q = Queue(1000)

    def __enter__(self):
        id = threading.get_ident()
        a = A(id)
        self._items[id] = a
        self.q.put(a)
        print("enter ", id, "size", len(self._items))
        return a

    def __exit__(self, exc_type, exc_val, exc_tb):
        id = threading.get_ident()
        print("Done Before", self._items)
        a = self._items.pop(id)
        s = random.randint(3, 10)
        out = self.q.get()
        sleep(s)
        if a is not None:
            a.quit()
        print("Done ", s, out)
        sleep(s)


from time import sleep
# from threading.pool import Pool
def test(mgr):
    sleep(random.random() + 0.1)
    with mgr as m:
        print("Currently using ", m.id)
        sleep(3)
    pass
if __name__ == "__main__":
    mgr = Mgr()
    threads = []
    for i in range(5):
        t = Thread(target=test, args=(mgr, ))
        t.start()
        threads.append(t)
    # print("====================\n\n")
    for i in threads:
        i.join()



import select
from bffacilities.utils import createLogger
logger = createLogger('network', savefile=False, stream=True, timeformat="%H:%M:%S")
import socket
from .sock import  SocketClient, UdpSocketClient, EventHandler
if hasattr(select, "epoll"):
    from .servers import BaseServer, TcpSocketServer, UdpSocketServer
else:
    from .socketserver_select import  BaseServer, TcpSocketServer, UdpSocketServer


# -*- coding: utf-8 -*-

"""
0.0.41: optmize network.socket 

0.0.40: move socket server and bluetooth into network, remove _globals

0.0.35: optimize scripts in torch folder

0.0.34: fix database error

0.0.33: move logger from utils to avoid log too much content

0.0.32: update with pyupdator

0.0.22: socket server has been improved a lot, interface has changed

v0.0.20 add bluetooth manager for bluetooth device list and connect

v0.0.19 some updates on generateQtTest and torch scripts

0.0.17: scripts in `myscripts` folder could add more scripts
without change main.py file.

0.0.13: add torch scripts and labelme scripts for MachineLearing data processing or model building.

0.0.7: add simple tcp socket server

0.0.6: add common flask functions
"""

__version__ = "0.0.41"
__author__ = "BriFuture"

from .utils import createLogger, initGetText
from ._constants import BFF_ROOT_PATH, BFF_OTHER_PATH, terminal_colors





"""some basic configurations
"""
from pathlib import Path
import os
import os.path as osp

_BASE_DIR = Path.home() / ".shells" # Basic Path
_CONFIG_DIR = _BASE_DIR / "configs"  # Default Config Path

_CONFIG_DIR.mkdir(parents=True, exist_ok=True)

BFF_ROOT_PATH = Path(osp.abspath(osp.dirname(__file__)))
BFF_OTHER_PATH = BFF_ROOT_PATH / "other"
BFF_TEMPLATE_PATH = BFF_ROOT_PATH / "template"

#Dictionary with console color codes to print text
terminal_colors = {
    'HEADER' : "\033[95m",
    'OKBLUE' : "\033[94m",
    'RED' : "\033[91m",
    'OKYELLOW' : "\033[93m",
    'GREEN' : "\033[92m",
    'LIGHTBLUE' : "\033[96m",
    'WARNING' : "\033[93m",
    'FAIL' : "\033[91m",
    'ENDC' : "\033[0m",
    'BOLD' : "\033[1m",
    'UNDERLINE' : "\033[4m" 
}


'''
 # @ Author: BriFuture
 # @ Create Time: 2021-08-19 12:14:59
 # @ Modified by: BriFuture
 # @ Modified time: 2021-10-21 13:48:15
 # @ Description:
 '''

import sys
import asyncio
from time import sleep
import threading, queue
import socket
from datetime import datetime
import json
__version__ = "0.1.0"
try:
    from bleak import BleakScanner, BleakClient
    if sys.platform == "win32":
        from Windows.Devices.Bluetooth.Advertisement import BluetoothLEAdvertisementFilter, BluetoothLEAdvertisement 
    else:
        raise ValueError("Unsupported")
except:
    raise ValueError("Please Install bleak first")

from bffacilities.utils import createLogger
import logging
logger = logging.getLogger("network")

class BluetoothWrapper(object):
    """封装蓝牙操作
    """
    IO_DATA_CHAR_UUID_W   = "0000ffe1-0000-1000-8000-00805f9b34fb"
    IO_DATA_CHAR_UUID_W2  = "0000ffe2-0000-1000-8000-00805f9b34fb"

    def __init__(self, **kwargs):
        self._btMsgProcessors = []
        processor = kwargs.get("btMsgProcessor")
        if processor is not None:
            self._btMsgProcessors.append(processor)
        self.loop = asyncio.get_event_loop()
        self._msgToBtQueue = queue.Queue()
        self._cmdQueue = queue.Queue()

        self._btClient = None
        self._btConnected = False
        self._devices = { } # key address
        self._lastAddress = None

    @property
    def is_connected(self):
        return self._btConnected
    @property
    def btMsgProcessor(self):
        return self._btMsgProcessors
    
    @btMsgProcessor.setter
    def btMsgProcessor(self, p):
        """
        Callback

            btMsgProcess(data: bytes)
        """
        self._btMsgProcessors.append(p)
    
    def export_deviceInfo(self):
        data = []
        for device in self._devices.values():
            name = device.name if hasattr(device, "name") else None
                
            data.append({
                "address": device.address,
                "name": name,
                "rssi": device.rssi
            })
        data = json.dumps(data)
        return data

    def syncDeviceToOther(self):
        """should be rewrite by subclass
        """
        pass

    def _onRecvBtMsg(self, sender: int, data: bytes):
        """
        Args
            data bytes
        """
        # logger.debug(f"Recv BT & Send: {len(data)}")
        for p in self._btMsgProcessors:
            p(data)

    def listDevice(self, timeout = 10):
        async def _list(timeout):
            devices = await BleakScanner.discover(timeout=timeout)
            for d in devices:
                if d.address not in self._devices:
                    self._devices[d.address] = d
            logger.info(f"{self._devices}")
            self.syncDeviceToOther()
        self.loop.run_until_complete(_list(timeout))
    
    def findDevice(self, name = None, seconds = 5.0):
        self.loop.run_until_complete(self.monitDevice(name, seconds))

    async def monitDevice(self, name, seconds=0):
        """
        Args:
            name  device name such as "BTDEVICE1"
        """
        bleFilter = BluetoothLEAdvertisementFilter()
        if name:
            ad = BluetoothLEAdvertisement()
            ad.put_LocalName(name)
            bleFilter.put_Advertisement(ad)

        scanner = BleakScanner(AdvertisementFilter=bleFilter)
        scanner.register_detection_callback(self._detection_callback)
        # await scanner.set_scanning_filter()
        await scanner.start()
        if seconds <= 5:
            while self._btConnected:
                await asyncio.sleep(10)
        else:
            await asyncio.sleep(seconds)
        await scanner.stop()
        logger.info("[BT] Monit Device Done")

    def _detection_callback(self, device, advertisement_data):
        if len(advertisement_data.local_name) == 0:
            return
        if device.address not in self._devices:
            logger.info(f"First [{datetime.now()}] RSSI: {device.rssi}, {advertisement_data.local_name}, {device.address}")
        device.name = advertisement_data.local_name
        self._devices[device.address] = device
        self.syncDeviceToOther()

    async def _connect(self, address):
        """after function exits, bluetooth device will be disconnected
        """
        if self._btClient is not None:
            await self._btClient.disconnect()

        self._btClient = BleakClient(address)
        logger.info(f"[BT] Connecting Bluetooth: {address}")
        # service = await self._btClient.get_services()
        # model_number = await self._btClient.read_gatt_char(BluetoothWrapper.IO_DATA_CHAR_UUID_W )
        # print(f"Model Number {model_number}: {''.join(map(chr, model_number))}")
        try:
            self._btConnected = await self._btClient.connect()
            logger.info(f"[BT] Client {self._btClient} connected: {self._btConnected}")
            if self._btConnected:
                await self._btClient.start_notify(\
                    BluetoothWrapper.IO_DATA_CHAR_UUID_W, self._onRecvBtMsg)
        except Exception as e:
            self._btConnected = False
            logger.warning(f"[BT] Connect failed: {e}")
        # finally:
            # await self._btClient.disconnect()


    async def _sendMsgToBt(self):
        if not self._msgToBtQueue.empty():
            sData = self._msgToBtQueue.get()
            try:
                await self._btClient.write_gatt_char(BluetoothWrapper.IO_DATA_CHAR_UUID_W, sData)
            except : 
                pass
            logger.debug(f"SendData Bt {sData}")
        else:
            await asyncio.sleep(0)
    
    def sendMsgToBt(self):
        self.loop.run_until_complete(self._sendMsgToBt())

    async def _connectAndListen(self, address):
        await self._connect(address)
        try:
            while self._btConnected:
                await self._sendMsgToBt()
        except Exception as e:
            self._btConnected = False
            logger.warning(f"[BT] Exception Listen: {e}")
        finally:
            await self._btClient.disconnect()
        logger.info("[BT] connect process done")

    def connectToBLE(self, address, listen=False):
        """
        Args: 
            address:  MAC address

            listen 
            if True, the event loop will run always and stop main thread

            if False, you have to call `sendMsgToBt` mannually if there is 
            pending msg needs to be sent to bluetooth

        Note:
            If a bluetooth is connected, it should be disconnect first
        """
        if self._btConnected:
            if address != self._lastAddress:
                self.disconnect()
            else:
                logger.warning("Bluetooth Device is already connected")
                return

        self._lastAddress = address
        if listen:
            self.loop.run_until_complete(self._connectAndListen(address))
        else:
            self.loop.run_until_complete(self._connect(address))

    def disconnect(self):
        if self._btConnected:
            logger.info("[BT] Stoping bluetooth, wait for a moment")
            async def _stop():
                await self._btClient.disconnect()
            self.loop.run_until_complete(_stop())
            logger.info("[BT] bluetooth stopped")
        else:
            logger.warning("[BT] Not Connected")


    def _processCmd(self, cmd):
        """
        {
            "action": "listDevice",
            "timeout": "60", // seconds
        }

        {
            "action": "findDevice",
            "name": "CMPS", // or no name for find all device
            "timeout": 30, // default 30
        }

        {
            "action": "disconnect" // disconnect if a device has connected
        }

        {
            "action": "quit" // quit program
        }
        """
        # logger.debug(f"Cmd { threading.get_ident()}")
        action = cmd["action"]
        if action == "connect":
            if "address" in cmd:
                self.connectToBLE(cmd["address"])
        elif action == "listDevice":
            self.listDevice(timeout= cmd.get("timeout", 10))
        elif action == "findDevice":
            self.findDevice(name=cmd.get("name"), seconds= cmd.get("timeout", 20))
        elif action == "disconnect":
            self.disconnect()
        elif action == "quit":
            self.disconnect()
            self._running = False
            logger.info("About to quit")
        else:
            logger.warning("Unsupported cmd", action)

    def processCmd(self):
        """注意，用于蓝牙接收和发送的操作，
        该方法应该在创建 BluetoothWrapper 的线程中调用
        """
        if not self._cmdQueue.empty():
            self._processCmd(self._cmdQueue.get()) 

from bffacilities.socketserver import UdpSocketServer

class BluetoothMgr(BluetoothWrapper) :
    """
    @See bffacilities.socketserver.UdpSocketServer

    默认绑定到 UDP 12001 地址上，用于接收其他模块的控制信息

    """

    def __init__(self, port = 12001, broadcast = True, **kwargs):
        super().__init__(**kwargs)
        self._bind = port
        self.sockServer = UdpSocketServer()
        self.sockServer.createListener(port = self._bind, ip="127.0.0.1")
        self.sockServer.addProcessor(self.recvCtlMsg)
        self._running = True
        if kwargs.get("btMsgProcessor") is None:
            self.btMsgProcessor = self.sockServer.send

    def transfer(self):
        self.btMsgProcessor = self.sockServer.send
    @property
    def is_running(self):
        return self._running

    def recvCtlMsg(self, msg: bytes, addr: tuple):
        # print(msg, addr)
        if msg.startswith(b"{"):
            try:
                cmd = json.loads(msg.decode())
                self._processCmd(cmd)
            except: pass
        else:
            self._msgToBtQueue.put(msg)

    def syncDeviceToOther(self):
        data = self.export_deviceInfo()
        self._sockSender.sendto(data.encode() + b"\n\n")

    def start(self):
        """start an loop in main thread 
        start another thread for udp server
        """
        try:
            while self._running:
                self.sendMsgToBt()
                self.sockServer.serve_once()
                sleep(0.1)
        except (KeyboardInterrupt, SystemExit):
            self.stop()
    
    start_one = start
    
    def stop(self):
        self.disconnect()
        self._running = False
        logger.info("==========Bt Mgr Quit===========")


if __name__ == "__main__":
    btMgr = BluetoothMgr()
    btMgr.start()


"""
Base Sock
"""
import socket
import logging
import traceback

__version__ = "0.1.0"

logger = logging.getLogger("network")
class EventHandler:
    def fileno(self):
        """Return the associated file descriptor"""
        raise ValueError('Not Implemented must implement')

    def handle_receive(self):
        'Perform the receive operation'
        pass

    def handle_send(self):
        'Send outgoing data'
        pass

    def handle_except(self):
        pass

    def is_valid(self):
        return True

class SocketClient(EventHandler):
    """
    1. 通过继承该类进行客户端编写，
    2. 通过 parser 回调函数进行消息读取

    Args:
        sock: socket
        address: client socket address recved from server,
            useful when other want to communicate or 
            identify the client

        except_callback: function(client: SocketClient)
    """
    def __init__(self, sock, address, **kwargs):
        self._sock = sock
        self._sock.setblocking(False)
        if address is None:
            try:
                self.address = self._sock.getsockname()
            except: 
                self.address = None
        else:
            self.address = address

        self._isValid = True

        self._parse = None
        self._server = None
        self.except_callback = kwargs.get("except_callback", None)
        self.bufferSize = kwargs.get("bufferSize", 2048)

    @staticmethod
    def createTcpClient(serverAddr):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sc = SocketClient(sock, None)
        sc._sock.setblocking(True)
        sc._server = serverAddr
        return sc

    @property
    def sock(self):
        return self._sock

    def connect(self):
        if self._server is None:
            return
        try:
            self._sock.connect(self._server)
            if self.address is None:
                self.address = self._sock.getsockname()
        except:
            logger.warning(f"[SC] Error connecting {self._server} [{self.address}]")
            self.handle_except()
        finally:
            pass

    def shutdown(self):
        self._sock.close()

    def __del__(self):
        self._sock.close()

    @property
    def parser(self):
        return self._parse

    @parser.setter
    def parser(self, parse):
        """传入一个解析器即可
        @deprecated use setParser
        ```py
        # data is a dict
        def parse(data, socketClient): print(data, socketClient)

        client.parser = parse
        ```
        """
        try:
            parse('""', self) # pass empty string
            self._parse = parse
        except:
            logger.critical(f"[TC] ***** Error testing parser, it will be ignored ****")

    def setParser(self, parser, testJson=True):
        if testJson:
            try:
                parser('""', self) # pass empty string
                self._parse = parser
            except:
                logger.critical(f"[TC] ***** Error testing parser, it will be ignored ****")
        else:
            self._parse = parser

    def is_valid(self):
        return self._isValid

    def fileno(self):
        return self._sock.fileno()

    def __repr__(self):
        return f'<Socket Client, {self._sock.fileno()}, {self.address}>'

    def handle_receive(self):
        """Handle receive data, default processor
        subclass could replace this methods to receive binary data
        """
        msg = None
        try:
            msg = self._sock.recv(self.bufferSize)
            if not msg:
                self.handle_except()
                return
        except socket.error:
            self.handle_except()
        except Exception as e:
            logger.warning(f"[TC] recv {e}")
        
        if msg is None: return

        if self._parse:
            try:
                self._parse(msg.decode(), self)
            except Exception as e:
                logger.warning(f"[TC] parse {msg}, {e}")
                traceback.print_exc()
        else:
            logger.debug(msg)

    def handle_send(self):
        pass

    def handle_except(self):
        """注意不要在 callback 中重新创建 client，否则函数调用堆栈将会溢出
        """
        self._sock.close()
        self._isValid = False
        if self.except_callback:
            self.except_callback(self)


class UdpSocketClient(EventHandler):
    def __init__(self, port, ip = "127.0.0.1", name="default", **kwargs):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.address = (ip, port)
        self.sock.settimeout(0)
        self.sock.setblocking(False)
        self.sock.bind(self.address)
        self.bufferSize = kwargs.get("bufferSize", 1024)
        self.name = name
        self._isValid = True
        self._parse = None
        self._peer = None
        self._peers = []

        self.except_callback = kwargs.get("except_callback", None)
        # self.keepFirstPeer = False # if true, the first peer will always be used

    @property
    def parser(self):
        return self._parse

    @parser.setter
    def parser(self, parse):
        """传入一个解析器即可
        ```
        def parse(data, recver): 
            print(data, "From", recver)

        client.parser = parse
        ```
        """
        self._parse = parse

    def send(self, msg):
        """
        Args
            msg bytes
        """
        for p in self._peers:
            self.sock.sendto(msg, p)

    def is_valid(self):
        return self._isValid

    def fileno(self):
        return self.sock.fileno()

    def __repr__(self):
        return f"<UdpSocketClient {self.name} {self.address} {self.bufferSize} {self.sock.fileno()}>"

    def handle_receive(self):
        """Handle receive data, default processor
        subclass could replace this methods to receive binary data
        """
        bytesAddressPair = None
        try:
            bytesAddressPair = self.sock.recvfrom(self.bufferSize)
        except Exception as e: 
            # logger.warning(f"[UC] {self}  recv error : {e}")
            # traceback.print_exc()
            # ICMP recved cause the error, ignore it
            pass

        # message = bytesAddressPair[0]
        if not bytesAddressPair:
            # self.handle_except()
            return
        if not bytesAddressPair[1] in self._peers:
            logger.info(f"[UC] new client added: {bytesAddressPair[1]}")
            self._peers.append(bytesAddressPair[1])
        if self._parse:
            try: 
                self._parse(bytesAddressPair, self.name)
            except Exception as e:
                logger.warning(f"[UC] parse error: {e}")
                traceback.print_exc()

    def handle_send(self):
        pass

    def handle_except(self):
        self.sock.close()
        self._isValid = False
        if self.except_callback:
            self.except_callback(self)

